<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 000ca24b71cb $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-06-25 17:20 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Class weTableItem
 *
 * Provides functions for exporting and importing table rows.
 */
class we_backup_tableItem extends we_backup_contentBase{
	//FIXME:can we add DELAYED INSERT and GROUP elements from one table?!

	public const BACKUP_XML_TAG = 'we:tableitem';

	/** @var string */
	var $table;

	/** @var string[] */
	static protected $tableKeys = [];

	/** @var array<string,mixed> */
	protected $elements = [];

	public function __construct(string $table, array $data = []){
		$this->table = $table;
		$this->elements = $data;
	}

	/**
	 * @staticvar array $cache
	 * @param string $table
	 * @return string[]
	 */
	static function getTableKey(string $table, bool $live = true): array{
		static $cache = [];
		if(isset($cache[$table])){
			return self::$tableKeys = $cache[$table];
		}
		//read Primary key from installed default files
		if(!$live && file_exists(WE_DOCUMENT_ROOT . '/webEdition/liveUpdate/sqldumps/' . stripTblPrefix($table) . '.sql')){
			$lines = we_base_file::loadLines(WE_DOCUMENT_ROOT . '/webEdition/liveUpdate/sqldumps/' . stripTblPrefix($table) . '.sql', 2, 2048);
			$cache[$table] = DB_WE::$inst->getPrimaryKeys($table, $lines);
		}
		if(!isset($cache[$table]) || !$cache[$table]){//fallback, or for external tables
			$cache[$table] = DB_WE::$inst->getPrimaryKeys($table);
		}
		return self::$tableKeys = $cache[$table];
	}

	/**
	 * @staticvar array $cache
	 * @param string $table
	 * @return string[]
	 */
	static function getGeneratedColumns(string $table, bool $live = true): array{
		static $cache = [];
		'@phan-var array<string,string[]> $cache';
		if(isset($cache[$table])){
			return $cache[$table];
		}
		//read Primary key from installed default files
		if(!$live && file_exists(WE_DOCUMENT_ROOT . '/webEdition/liveUpdate/sqldumps/' . stripTblPrefix($table) . '.sql')){
			$lines = we_base_file::loadLines(WE_DOCUMENT_ROOT . '/webEdition/liveUpdate/sqldumps/' . stripTblPrefix($table) . '.sql', 2, 2048);
			$cache[$table] = array_keys(DB_WE::$inst->getGeneratedColumns($table, $lines));
		}
		if(!isset($cache[$table]) || !$cache[$table]){//fallback, or for external tables
			$cache[$table] = array_keys(DB_WE::$inst->getGeneratedColumns($table));
		}
		return $cache[$table];
	}

	/**
	 *
	 * @staticvar array $cache
	 * @param string $table
	 * @return array<string,mixed>
	 */
	protected static function getDefaults(string $table): array{
		static $cache = [];
		'@phan-var array<string,array<string,mixed>> $cache';
		if(isset($cache[$table])){
			return $cache[$table];
		}
		$data = [];
		foreach(DB_WE::$inst->metadata($table, we_database_base::META_FULL) as $name => $item){
			if(isset($item['default'])){
				$data[$name] = $item['default'];
			}
		}
		$cache[$table] = $data;
		return $cache[$table];
	}

	/**
	 *
	 * @staticvar array $cache
	 * @param string $table
	 * @return string[]
	 */
	protected static function getBinFields(string $table): array{
		static $cache = [];
		'@phan-var array<string,string[]> $cache';
		if(isset($cache[$table])){
			return $cache[$table];
		}
		$data = [];
		foreach(DB_WE::$inst->metadata($table, we_database_base::META_FULL) as $name => $item){
			if($item['type_short'] === we_database_base::TYPE_SHORT_BINARY){
				$data[] = $name;
			}
		}
		$cache[$table] = $data;
		return $cache[$table];
	}

	function getFieldType(string $fieldname): string{
		$regs = [];
		return (preg_match('/(.+?)_(.*)/', $fieldname, $regs) ? $regs[1] : '');
	}

	public function getLogString(string $prefix = ''): string{
		$id_val = $prefix . $this->table . '(' . implode(',', self::$tableKeys) . ')';

		foreach(self::$tableKeys as $key){
			$id_val .= ':' . ($key === 'nHash' ? bin2hex($this->elements[$key]) : $this->elements[$key]);
		}
		return $id_val;
	}

	public function getContentType(): string{
		return $this->elements['ContentType'] ?? '';
	}

	public function getID(): int{
		return (int) ($this->elements['ID'] ?? 0);
	}

	/**
	 * @return array{elements:array}
	 */
	protected function getBackupData(): array{
		foreach(self::getBinFields($this->table) as $name){
			if($this->elements[$name] !== null){
				$this->elements[$name] = bin2hex($this->elements[$name]);
			}
		}
		$defaults = self::getDefaults($this->table);
		//skip null values, these are defaults; save some space
		foreach($this->elements as $name => $el){
			if((isset($defaults[$name]) && $el === $defaults[$name]) ||
				$el === '0000-00-00' ||
				$el === '0000-00-00 00:00:00'){
				unset($this->elements[$name]);
			}
		}
		return ['elements' => array_diff_key($this->elements, array_flip(self::getGeneratedColumns($this->table, false)))];
	}

	public function setBackupData(array $data): bool{
		$this->elements = $data['elements'];
		unset($data);
		$genCols = self::getGeneratedColumns($this->table, false);
		foreach(self::getBinFields($this->table) as $name){
			if(in_array($name, $genCols)){
				continue;
			}
			$this->elements[$name] = $this->elements[$name] === null ? null : hex2bin($this->elements[$name]);
		}
//only set elements, which are not generated; compare to live table!
		$this->elements = array_diff_key($this->elements, array_flip($genCols));
		foreach($this->elements as $name => $el){
			if($el === '0000-00-00' ||
				$el === '0000-00-00 00:00:00'){
				unset($this->elements[$name]);
			}
		}
		return DB_WE::$inst->query('INSERT IGNORE INTO ' . $this->table . ' SET ' . we_database_base::arraySetter($this->elements));
	}

	protected function getTable(): string{
		return we_backup_util::getDefaultTableName($this->table);
	}

}
