<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 735d895ec82e $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2022-03-02 16:32 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * General Definition of WebEdition Customer
 *
 */
class we_customer_customer extends we_base_model{
	const NOPWD_CHANGE = '__WE__PWD_NO_CHANGE';
	const ENCRYPTED_DATA = '**ENCRYPTED**';
	public const ENCRYPT_SYMMETRIC = 1;
	public const ENCRYPT_HASH = 2;
	public const REMOVE_PASSWORD = 0;
	public const STORE_PASSWORD = 1;
	public const STORE_DBPASSWORD = 2;
	public const PWD_ALL_OK = 0;
	public const PWD_FIELD_NOT_SET = 1;
	public const PWD_NOT_MATCH = 2;
	public const PWD_NO_SUCH_USER = 3;
	public const PWD_TOKEN_INVALID = 4;
	public const PWD_USER_EMPTY = 5;
	public const PWD_USER_EXISTS = 6;
	public const PWD_NOT_SUFFICIENT = 7;
	public const PWD_UNKNOWN_ERROR = 8;
	public const PWD_USER_NOT_UNIQUE = 9;
	const CUSTOMER_AUTOLOGIN_LIFETIME = 31536000/* 365*24*3600 */;
	const PROPERTY_FIELDS = [
		'AutoLogin',
		'AutoLoginDenied',
		'Forename',
		'LastAccess_Date',
		'LastLogin_Date',
		'LoginDenied',
		'MemberSince_Date',
		'Password',
		'Surname',
		'Username',
	];
	const PROTECTED_FIELDS = [
		'ID',
		'LastAccess',
		'LastAccess_Date',
		'LastLogin',
		'LastLogin_Date',
		'MemberSince',
		'MemberSince_Date',
		'Mod_Date',
		'ModifiedBy',
		'ModifyDate',
	];
	const IGNORED_EXPORT_FIELDS = [
		'LastAccess',
		'LastAccess_Date',
		'LastLogin',
		'LastLogin_Date',
		'MemberSince',
		'MemberSince_Date',
		'Mod_Date',
		'ModifyDate',
		'Password',
	];

	//properties
	/** @var int */
	var $ID = 0;

	/** @var string */
	var $Username;

	/** @var string */
	var $Password;

	/** @var bool */
	var $LoginDenied;

	/** @var string */
	var $Forename;

	/** @var string */
	var $Surname;

	/** @var DateTime */
	var $MemberSince_Date;

	/** @var ?DateTime */
	var $LastLogin_Date = null;

	/** @var ?DateTime */
	var $LastAccess_Date = null;

	/** @var DateTime */
	var $Mod_Date;

	/** @var string */
	var $ModifiedBy = 'backend';

	/** @var bool */
	var $AutoLoginDenied = false;

	/** @var bool */
	var $AutoLogin = false;

	/** @var array<string,array> */
	private static $fieldAdds = [];

	/** @var string[] */
	protected static $fieldsNull = [];

	/** @var array<string,string> */
	protected static $fieldsReq = [];

	/** @var string[] this contains the names of required fields without input or defaults */
	protected $requiredEmpty = [];


	/**
	 * Default Constructor
	 * Can load or create new Customer depends of parameter
	 */
	function __construct(int $customerID = 0){
		parent::__construct(CUSTOMER_TABLE);
		$this->loadPresistents();
		if($customerID){
			$this->load($customerID);
			return;
		}
		$this->MemberSince_Date = new DateTime();
	}

	public function reloadPersistenSlots(): void{
		$this->loadPresistents(true);
	}

	protected function loadPresistents(bool $force = false): void{
		if(!$force && $this->table === self::$persistentTable){
			$this->persistent_slots = self::$persistentSlots;
			return;
		}
		self::$persistentTable = $this->table;

		$this->persistent_slots = [];
		// we need the default values from table metadata (META_FULL)
		$tableInfo = $this->db->metadata($this->table, we_database_base::META_FULL);
		$generated = array_keys($this->db->getGeneratedColumns($this->table));

		foreach($tableInfo as $info){
			$fname = (string) $info['name'];
			if(in_array($fname, $generated)){
				continue;
			}

			if(!($info['flags'] & MYSQLI_NOT_NULL_FLAG)){
				self::$fieldsNull[] = $fname;
			}

			$specialDefault = NULL;
			switch($info['type']){
				case we_database_base::TYPE_TINYBLOB:
				case we_database_base::TYPE_MEDIUMBLOB:
				case we_database_base::TYPE_BLOB:
				case we_database_base::TYPE_LONGBLOB:
				case we_database_base::TYPE_VARBIN:
				case we_database_base::TYPE_BIN:
					$this->binFields[] = $fname;
					$this->persistent_slots[$fname] = we_base_request::RAW;
					break;
				case we_database_base::TYPE_TIMESTAMP:
				case we_database_base::TYPE_DATE:
				case we_database_base::TYPE_DATETIME:
					$this->dateFields[] = $fname;
					$this->persistent_slots[$fname] = we_base_request::DATE_TIME;
					break;
				case we_database_base::TYPE_TINYINT:
				case we_database_base::TYPE_SMALLINT:
				case we_database_base::TYPE_INT:
				case we_database_base::TYPE_MEDIUMINT:
				case we_database_base::TYPE_BIGINT:
					$this->persistent_slots[$fname] = we_base_request::INT;
					break;
				case we_database_base::TYPE_FLOAT:
				case we_database_base::TYPE_DOUBLE:
				case we_database_base::TYPE_DECIMAL:
					$this->persistent_slots[$fname] = we_base_request::FLOAT;
					break;
				case we_database_base::TYPE_SET:
					$this->persistent_slots[$fname] = we_base_request::STRING_LIST;
					break;
				case we_database_base::TYPE_ENUM:
					$specialDefault = in_array('',$info['values'], true) ? '' : NULL;
				default:
					$this->persistent_slots[$fname] = we_base_request::STRING;
			}

						if(($info['flags'] & MYSQLI_NOT_NULL_FLAG) &&
							!in_array($fname, [
								'ID',
								'MemberSince_Date',
								'Mod_date',
								'ModifiedBy',
								'Password',
								'Forename',
								'Surname',
								'LoginDenied',
								'AutoLogin',
								'AutoLoginDenied',
								], true)){
				self::$fieldsReq[$fname] = $info['default'] ?? $specialDefault;
			}

		}
		self::$persistentSlots = $this->persistent_slots;
	}

	/**
	 * 	Overwrites an existing User, keeps customerfilters => User should be the same person
	 * @param int $id id to overwrite
	 */
	function overwrite(int $id): bool{
		$tmp = new self($id);
		$tmp->delete(false);
		unset($tmp);
		$this->ID = $id;
		return $this->save();
	}

	function save(bool $force_new = false, bool $rebuild = false, array $skipWrite = []): bool{
		if(!$this->MemberSince_Date){
			$this->MemberSince_Date = new DateTime();
		}
		$this->Mod_Date = new DateTime();
		$this->ModifiedBy = 'backend';

		foreach(self::$fieldsReq as $nField => $default){
			// check if there is meaningful input of required type
			// fixme: we should check for changes to original data
			$input = we_base_request::_($this->persistent_slots[$nField], 'customer', NULL, $nField) ?: ($this->{$nField} ?? ($this->{$nField} = $default));

			if(!$input && is_null($default)){
				$this->requiredEmpty[$nField] = $nField;
			} else {
				unset($this->requiredEmpty[$nField]);
			}
		}
		if($this->requiredEmpty){
			return false;
		}
		foreach(self::$fieldsNull as $nField){
			if(empty($this->{$nField})){
				$this->{$nField} = NULL;
			}
		}
		$hook = new we_hook_handler('customer_preSave', ['customer' => $this, 'from' => 'management', 'type' => ($this->ID ? 'existing' : 'new')]);

		return $hook->executeHook() && we_base_model::save() && $this->registerMediaLinks();
	}

	function getRequiredEmpty():array{
		return $this->requiredEmpty;
	}

	function registerMediaLinks(): bool{
		$this->unregisterMediaLinks();
		foreach(self::getImageFields() as $field){
			if($this->{$field}){
				$this->MediaLinks[$field] = $this->{$field};
			}
		}

		parent::registerMediaLinks();
		return true;
	}

	/**
	 * delete entry from database
	 * @param $recursive if true, customerfilter are deleted as well
	 */
	function delete(bool $recursive = true): bool{ //FIXME: what about documents/objects of customer?
		if(we_base_model::delete() && $recursive){
			we_customer_abstractFilter::deleteWebUser($this->ID);
			return true;
		}
		return false;
	}

	static function transFieldName(string $real_name, string &$branch): string{
		// Fixme: what is this supposed to do?
		if(str_contains($real_name, g_l('modules_customer', '[other]'))){
			return $real_name;
		}
		$pre = explode('_', $real_name);
		if(($pre[0] != $real_name) && (!in_array($pre[0], self::PROTECTED_FIELDS, true)) && (!in_array($pre[0], self::PROPERTY_FIELDS, true))){
			$branch = $pre[0];
			$field = implode('_', array_slice($pre, 1));
			return $field;
		}
		return $real_name;
	}

	/**
	 *
	 * @return array{branches:array<string,array<string,mixed>>,fixed:array<string,mixed>,other:array<string,mixed>}
	 */
	function getBranches(): array{
		$other = $branches = $fixed = [];
		$fixed['ID'] = $this->ID; // Bug Fix #8413 + #8520

		$branch = '';
		foreach(array_keys($this->persistent_slots) as $per){
			$per = (string) $per;
			$var_value = ($this->{$per} ?? null);
			$field = self::transFieldName($per, $branch);

			if(in_array($per, self::PROPERTY_FIELDS, true)){
				$fixed[$per] = $var_value;
			} else if(in_array($per, self::PROTECTED_FIELDS, true)){
				continue;
			} elseif($field != $per){
				$branches[$branch][$field] = $var_value;
			} else {
				$other[$per] = $var_value;
			}
		}
		ksort($branches);
		return ['branches' => $branches, 'fixed' => $fixed, 'other' => $other];
	}

	/**
	 * @return string[]
	 */
	function getBranchesNames(): array{
		['branches' => $branches] = $this->getBranches();
		return array_keys($branches);
	}

	/**
	 * @param string $branch
	 * @return string[]
	 */
	function getFieldsNames(string $branch): array{
		$ret = $arr = [];
		['branches' => $branches, 'fixed' => $common, 'other' => $other] = $this->getBranches();

		$branch = $branch ?: g_l('modules_customer', '[other]');

		switch($branch){
			case g_l('modules_customer', '[common]'):
				if(is_array($common)){
					$arr = $common;
				}
				break;
			case g_l('modules_customer', '[other]'):
				if(is_array($common)){
					$arr = $other;
				}
				break;
			default:
				if(isset($branches[$branch]) && is_array($branches[$branch])){
					$arr = $branches[$branch];
				}
		}

		$other = g_l('modules_customer', '[other]');
		foreach(array_keys($arr) as $b){
			if($branch === $other){
				$ret[$b] = $b;
			} else {
				$ret[$branch . '_' . $b] = $b;
			}
		}
		return $ret;
	}

	static function isProperty(string $field): bool{
		return in_array($field, self::PROPERTY_FIELDS, true);
	}

	static function isProtected(string $field): bool{
		return in_array($field, self::PROTECTED_FIELDS, true);
	}

	function clearSessionVars(): void{
		if(isset($_SESSION['weS']['module_sessions']['customer'])){
			unset($_SESSION['weS']['module_sessions']['customer']);
		}
	}

	static function customerNameExist(string $name): bool{
		$name = trim($name);
		return ($name ? fB('SELECT 1 FROM ' . CUSTOMER_TABLE . ' WHERE Username="' . DB_WE::$inst->escape($name) . '" LIMIT 1') : true);
	}

	public function isDuplicateUsername(): int{
		return fI('SELECT ID FROM ' . CUSTOMER_TABLE . ' WHERE Username="' . DB_WE::$inst->escape($this->Username) . '" AND ID!=' . intval($this->ID));
	}

	function fieldExist(string $field): bool{
		return isset($this->persistent_slots[$field]);
	}

	/**
	 *
	 * @return string[]
	 */
	public static function getFieldset(): array{
		$result = [];
		$fields = DB_WE::$inst->metaFieldNames(CUSTOMER_TABLE);
		foreach($fields as $k){
			if(!self::isProtected((string) $k)){
				$result[] = $k;
			}
		}
		return $result;
	}

	function filenameNotValid(): bool{
		return preg_match('|[/]|i', $this->Username) ? true : false;
	}

	public static function cryptPassword(string $pass): string{
		return we_users_user::makeSaltedPassword(substr($pass, 0, we_users_user::MAX_PWD_LEN), 10);
		//}
	}

	public static function comparePassword(string $storedPassword, string $clearPassword, int $uid): bool{
		if(!$storedPassword || !$clearPassword){
			return false;
		}
		$clearPassword = substr($clearPassword, 0, we_users_user::MAX_PWD_LEN);
		$matches = [];
		if(!preg_match('|^\$([^$]{2,4})\$([^$]+)\$(.+)$|', $storedPassword, $matches)){
			$ret = $storedPassword === $clearPassword;
			if($ret){
				DB_WE::$inst->query('UPDATE ' . CUSTOMER_TABLE . ' SET Mod_Date=Mod_Date,Password="' . self::cryptPassword($clearPassword) . '" WHERE ID=' . $uid);
			}
			return $ret;
		}
		$latin = mb_convert_encoding($clearPassword, 'ISO-8859-15', 'UTF-8') ?: '';
		switch($matches[1]){
			case '-1':
				$ret = ($clearPassword === self::decryptData($storedPassword));
				if($ret){
					DB_WE::$inst->query('UPDATE ' . CUSTOMER_TABLE . ' SET Mod_Date=Mod_Date,Password="' . self::cryptPassword($clearPassword) . '" WHERE ID=' . $uid);
				}
				return $ret;
			case '2y':
				$ret = we_users_user::comparePasswords($storedPassword, $clearPassword);
				if(!$ret && $latin && $latin != $clearPassword){
					$ret = we_users_user::comparePasswords($storedPassword, $latin);
					if($ret){
						DB_WE::$inst->query('UPDATE ' . CUSTOMER_TABLE . ' SET Mod_Date=Mod_Date,Password="' . self::cryptPassword($clearPassword) . '" WHERE ID=' . $uid);
					}
				}
				return $ret;
		}
		return false;
	}

	public static function cryptGetIV($lenMethod): string{
		$len = is_numeric($lenMethod) ? (int) $lenMethod : (int) openssl_cipher_iv_length($lenMethod);
		return random_bytes($len);
	}

	public static function cryptData(string $data, string $key = SECURITY_ENCRYPTION_KEY, bool $keepBin = false): string{//Note we need 4 Bytes prefix + 16 Byte IV + 1$ = 21 Bytes. The rest is avail for data, which is hex'ed, so "half" of length is available
		if($data && function_exists('openssl_encrypt')){
			$iv = self::cryptGetIV('blowfish');
			$data = (string) openssl_encrypt($data, 'blowfish', hex2bin($key) ?: '', OPENSSL_RAW_DATA, $iv);
			return '$-1' . ($keepBin ? 'a' : '') . '$' . bin2hex($iv) . '$' . ($keepBin ? $data : (bin2hex($data) ?: ''));
		}
		return $data;
	}

	public static function decryptData(string $data, string $key = SECURITY_ENCRYPTION_KEY): string{
		$matches = [];
		if(!preg_match('|^\$([^$]{2,4})\$([a-f0-9]{16,32})\$|i', $data, $matches)){
			return '';
		}
		$data = (string) substr($data, strlen($matches[0]));
		switch($matches[1]){
			case '-1':
				$data = hex2bin($data) ?: '';
			case '-1a'://@phan-suppress-current-line PhanPluginDuplicateSwitchCaseLooseEquality
				if(function_exists('openssl_decrypt')){
					return openssl_decrypt($data, 'blowfish', hex2bin($key) ?: '', OPENSSL_RAW_DATA, hex2bin($matches[2]) ?: '') ?: '';
				}
				return '';
			default:
			case '2y'://can't be decoded
				return '';
		}
	}

	public static function getEncryptedFields(): array{
		static $fields = false;
		if(is_array($fields)){
			return $fields;
		}
		self::$fieldAdds = self::$fieldAdds ?: we_unserialize(fS('SELECT pref_value FROM ' . SETTINGS_TABLE . ' WHERE tool="customer" AND pref_name="FieldAdds"'));
		$fields = [];

		foreach(self::$fieldAdds as $key => $value){
			if(!empty($value['encryption'])){
				$fields[$key] = self::ENCRYPTED_DATA;
			}
		}

		return $fields;
	}

	public static function getImageFields(): array{
		static $fields = -1;
		if(is_array($fields)){
			return $fields;
		}
		self::$fieldAdds = self::$fieldAdds ?: we_unserialize(fS('SELECT pref_value FROM ' . SETTINGS_TABLE . ' WHERE tool="customer" AND pref_name="FieldAdds"'));
		if(!self::$fieldAdds){
			return [];
		}

		$fields = [];
		foreach(self::$fieldAdds as $key => $value){
			if(!empty($value['type']) && $value['type'] === 'img'){
				$fields[] = $key;
			}
		}

		return $fields;
	}

	public static function getJSLangConsts(): string{
		return we_base_util::setLangString([
				'view' => [
					'delete_alert' => g_l('modules_customer', '[delete_alert]'),
					'nothing_to_delete' => (g_l('modules_customer', '[nothing_to_delete]')),
					'nothing_to_save' => (g_l('modules_customer', '[nothing_to_save]')),
					'reset_failed_login_successfully' => (g_l('modules_customer', '[login_reset_ok]')),
					'save_changed_customer' => g_l('modules_customer', '[save_changed_customer]'),
				],
				'admin' => [
					'add_field' => g_l('modules_customer', '[add_field]'),
					'branch_no_edit' => (g_l('modules_customer', '[branch_no_edit]')),
					'del_fild_question' => g_l('modules_customer', '[del_fild_question]'),
					'edit_branch' => g_l('modules_customer', '[edit_branche]'),
					'edit_field' => g_l('modules_customer', '[edit_field]'),
					'no_branch' => (g_l('modules_customer', '[no_branch]')),
					'no_field' => (g_l('modules_customer', '[no_field]')),
					'other' => g_l('modules_customer', '[other]'),
					'we_fieldname_notValid' => (g_l('modules_customer', '[we_fieldname_notValid]')),
					'select_duplicate_exists' => g_l('modules_customer', '[select_duplicate_exists]'),
					'select_is_not_subset' => g_l('modules_customer', '[select_is_not_subset]'),
				],
				'sortAdmin' => [
					'default_soting_no_del' => (g_l('modules_customer', '[default_soting_no_del]')),
					'sortname_empty' => (g_l('modules_customer', '[sortname_empty]')),
				]
		]);
	}

	public function setFromArray(array $data): void{
		foreach($this->persistent_slots as $name => $type){
			if(in_array($name, self::IGNORED_EXPORT_FIELDS) ||
				!isset($data[$name])
			){
				continue;
			}

			//@phan-suppress-next-line PhanDeprecatedFunction
			$this->{$name} = we_base_request::filterVar($data[$name], $type);
		}
	}

}
