<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 967c728473a5 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2021-07-01 17:24 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_exim_Export extends we_exim_ExIm{
	/** @var we_database_base */
	protected $db;

	/** @var string */
	protected $exportType;

	/** @var array<string,string> */
	protected $exportProperties = [];

	/** @var string[] */
	protected $permittedContentTypes = [];

	/** @var resource|false */
	protected $fh;

	/** @var bool */
	protected $export_binary = true;

	const EXPORT_PRE_PROCESS = 'initExport';
	const EXPORT_PROCESS_NEXT = 'writeNextItem';
	const EXPORT_POST_PROCESS = 'finishExport';
	const SELECTIONTYPE_CLASSNAME = 'classname';
	const SELECTIONTYPE_DOCTYPE = 'doctype';
	const SELECTIONTYPE_DOCUMENT = 'document';

	abstract protected function writeExportItem(we_base_mainmodel $doc, array $attribute = []): void;

	abstract protected function fileComplete(): void;

	function __construct(){
		parent::__construct();
		$this->db = new DB_WE;
	}

	public function getNextTask(): string{
		return !self::isInitialized() ? self::EXPORT_PRE_PROCESS :
			($this->RefTable->current !== $this->RefTable->getCount() ? self::EXPORT_PROCESS_NEXT : self::EXPORT_POST_PROCESS);
	}

	public function exportPreprocess(we_export_export $export): void{
		$this->initExporter($export);
		$this->prepareExport();
		$this->savePreserves();
	}

	public function exportPostprocess(): void{
		$this->fileComplete();
		self::unsetPreserves();
	}

	public function exportNext(): array{
		$all = $this->RefTable->getCount();
		$ref = $this->RefTable->getNext();
		$this->savePreserves();

		if(!$ref->ID || !$ref->ContentType || !$ref->Table){
			return ['success' => false];
		}

		$table = $this->db->escape($ref->Table);
		if($ref->ContentType === we_base_ContentTypes::BACKUP_BINARY || fB('SELECT 1 FROM ' . $table . ' WHERE ID=' . intval($ref->ID))){
			$current = $this->RefTable->current;
			$this->export($ref->ID, $ref->ContentType, $table);

			return ['success' => true, 'id' => $ref->ID, 'contentType' => $ref->ContentType, 'table' => $table, 'current' => $current, 'total' => $all];
		}

		return ['success' => false];
	}

	protected function initExporter(we_export_export $export): void{
		self::unsetPreserves();
		$finalDocs = [];
		$finalTempl = [];
		$finalObjs = [];
		$finalClasses = [];
		$handle_documents = 0;
		$handle_templates = 0;
		$handle_classes = 0;
		$handle_objects = 0;

		// FIXME: this may be obsolete when $xmlExIm->getSelectedItems() works correctly
		if($export->Selection === 'manual'){
			switch($export->ExportType){
				case we_exim_ExIm::TYPE_CSV:
				case we_exim_ExIm::TYPE_XML:
					if($export->XMLTable === stripTblPrefix(FILE_TABLE)){
						$finalDocs = $export->selDocs;
						$handle_documents = 1;
					} else {
						$finalObjs = $export->selObjs;
						$handle_objects = 1;
					}
					$export->ExportDepth = 1;
					break;
				default:
					$finalDocs = $export->selDocs;
					$finalTempl = $export->selTempl;
					$finalObjs = $export->selObjs;
					$finalClasses = $export->selClasses;
					$handle_documents = 1;
					$handle_templates = 1;
					$handle_classes = 1;
					$handle_objects = 1;
			}
		}

		$this->getSelectedItems($export->Selection, $export->ExportType, $export->XMLTable, $export->SelectionType, $export->DocType, $export->TableID, $export->Categorys, $export->Folder, $finalDocs, $finalTempl, $finalObjs, $finalClasses);

		$this->setOptions([
			'export_type' => $export->ExportType,
			'handle_def_templates' => $export->HandleDefTemplates,
			'handle_doctypes' => $export->HandleDoctypes,
			'handle_categorys' => $export->HandleCategorys,
			'handle_def_classes' => $export->HandleDefClasses,
			'handle_document_includes' => $export->HandleDocIncludes,
			'handle_document_linked' => $export->HandleDocLinked,
			'handle_object_includes' => $export->HandleObjIncludes,
			'handle_object_embeds' => $export->HandleObjEmbeds,
			'handle_class_defs' => $export->HandleDefClasses,
			'handle_owners' => $export->HandleOwners,
			'export_depth' => $export->ExportDepth,
			'handle_documents' => $handle_documents,
			'handle_templates' => $handle_templates,
			'handle_classes' => $handle_classes,
			'handle_objects' => $handle_objects,
			'handle_navigation' => $export->HandleNavigation,
			'handle_thumbnails' => $export->HandleThumbnails,
			'xml_cdata' => $export->XMLCdata,
			'xml_table' => $export->XMLTable,
			'csv_delimiter' => $export->CSVDelimiter,
			'csv_lineend' => $export->CSVLineend,
			'csv_enclose' => $export->CSVEnclose,
			'csv_fieldnames' => $export->CSVFieldnames
		]);

		$this->exportProperties = [
			'exportType' => $this->exportType,
			'fileName' => $export->Filename,
			'file' => $export->ExportFilename
		];

		$this->RefTable->reset();
		$this->savePreserves();
	}

	public function prepareExport(array $ids = []): void{ // check access level of parent
		$this->fileCreate();
		$this->RefTable->reset();

		if($this->options['handle_owners']){
			we_base_file::save($this->exportProperties['file'], self::exportInfoMap($this->RefTable->Users), "ab");
		}
	}

	protected function fileCreate(): bool{
		if(empty($this->exportProperties['file'])){
			return false;
		}

		return (file_exists($this->exportProperties['file']) ?
			we_base_file::deleteLocalFile($this->exportProperties['file']) :
			true);
	}

	protected function setRefTable(&$rTable): void{
		$this->RefTable = $rTable;
	}

	protected function export(int $id, string $ct, string $table = ''): bool{
		$doc = we_exim_contentProvider::getInstance($ct, $id, $table);
		assert($doc instanceof we_document_base || $doc instanceof we_backup_contentBase);

		// add binary data separately to stay compatible with the new binary feature in v5.1
		if(in_array($doc->ContentType, [we_base_ContentTypes::IMAGE, we_base_ContentTypes::VIDEO, we_base_ContentTypes::AUDIO, we_base_ContentTypes::PDF], true)){
			assert($doc instanceof we_document_binary);
			//FIXME: this should be determined by document, not by type
			$doc->setElementS('data', we_base_file::load(WE_DOCUMENT_ROOT . WE_SITE_DIR . $doc->Path));
		}

		$this->fh = fopen($this->exportProperties['file'], 'ab');
		if(!$this->fh){
			return false;
		}

		$params = [];
		if(isset($doc->ID)){
			$params['ID'] = $doc->ID;
		}
		if(isset($doc->ContentType)){
			$params['ContentType'] = $doc->ContentType;
		} elseif(isset($doc->Table) && $doc->Table === DOC_TYPES_TABLE){
			$params['ContentType'] = 'doctype';
		}

		$this->RefTable->setProp($params, 'Examined', true);

		$tag = constant(get_class($doc) . '::BACKUP_XML_TAG');

		$attribute = ($doc instanceof we_backup_contentBase ? $doc->attribute_slots : []);

		switch($tag){
			case we_backup_dbTable::BACKUP_XML_TAG:
				assert($doc instanceof we_backup_dbTable);
				if(str_starts_with(strtolower($doc->table), strtolower(stripTblPrefix(OBJECT_X_TABLE))) ||
					we_modules_info::isActive(we_modules_info::CUSTOMER)){
					$doc->getColumns();
				}
			//no break;
			default:
				$this->writeExportItem($doc, $attribute);
				break;
		}

		if($this->export_binary &&
			$tag === we_backup_tableItem::BACKUP_XML_TAG &&
			strtolower($doc->table) === strtolower(FILE_TABLE) &&
			(in_array($doc->ContentType, [we_base_ContentTypes::IMAGE, we_base_ContentTypes::VIDEO, we_base_ContentTypes::AUDIO, we_base_ContentTypes::PDF], true))){
			$bin = we_exim_contentProvider::getInstance(we_base_ContentTypes::BACKUP_BINARY, $doc->ID);
			assert($bin instanceof we_backup_binary);
			//$attribute = ($bin->attribute_slots ?? []);
			$bin->binary2file($this->fh);
		}

		if($this->fh){
			fclose($this->fh);
			$this->fh = false;
		}

		unset($doc);
		return true;
	}

	// FIXME: move this to some new class we_exim_contentProviderNonWE
	protected function document2nonWE(we_contents_root $we_doc, bool $cdata = false): bool{
		if($we_doc->ContentType !== we_base_ContentTypes::WEDOCUMENT || !$this->fh){
			return false;
		}
		assert($we_doc instanceof we_document_webEdition);

		$template_code = fS('SELECT c.Dat FROM ' . CONTENT_TABLE . ' c WHERE c.DocumentTable="' . stripTblPrefix(TEMPLATES_TABLE) . '" AND c.DID=' . intval($we_doc->TemplateID) . ' AND c.nHash=unhex(md5("' . we_contents_template::FULLCODE . '"))');
		$tag_parser = new we_weTag_parser($template_code);
		$tags = $tag_parser->getAllTags();
		$regs = $records = [];

		foreach($tags as $tag){
			'@phan-var string $tag';
			if(!preg_match('|<we:([^> /]+)|i', $tag, $regs)){
				continue;
			}
			$tag_name = $regs[1];
			switch($tag_name){
				case 'var':
					continue 2;
				default:
					if(!preg_match('|name="([^"]+)"|i', $tag, $regs)){
						continue 2;
					}
					$name = $regs[1];
					switch($tag_name){
						// tags with text content, links and hrefs
						case 'input':
						case 'textarea':
						case 'href':
						case 'link':
							$records[] = $name;
							break 2;
					}
			}
		}

		$hrefs = [];
		$content = '';
		$tag_counter = 0;
		$usedFields = [];

		foreach($we_doc->elements as $k => $v){
			$tag_counter++;

			switch($v['type'] ?? ''){
				case 'date': // is a date field
					$tag_name = self::correctTagname($k, 'date', $tag_counter);

					$value = $we_doc->elements[$k]['dat'] ?: ($we_doc->elements[$k][we_contents_base::ETYPE_NUMBER] ?: 0); // FIXME: dates must not be stored in bdid!

					$content .= $this->formatOutput((string) ((int) $value), $tag_name, $this->exportType, 2, $cdata);
					$usedFields[] = $tag_name;

					// Remove tagname from array
					if($records){
						$records = $this->remove_from_check_array($records, $tag_name);
					}

					break;
				case 'txt':
					if(preg_match('|(.+)' . we_base_link::MAGIC_INFIX . '(.+)|', $k, $regs)){ // is a we:href field
						if(!in_array($regs[1], $hrefs, true)){
							$hrefs[] = $regs[1];

							if($we_doc->getElementI($regs[1] . we_base_link::MAGIC_INT_LINK)){
								$intID = $we_doc->getElementI($regs[1] . we_base_link::MAGIC_INT_LINK_ID);

								$tag_name = self::correctTagname($k, 'link', $tag_counter);
								$content .= $this->formatOutput(id2path($intID, FILE_TABLE), $tag_name, $this->exportType, 2, $cdata);
								$usedFields[] = $tag_name;

								// Remove tagname from array
								if($records){
									$records = $this->remove_from_check_array($records, $tag_name);
								}
							} else {
								$tag_name = self::correctTagname($k, 'link', $tag_counter);
								$content .= $this->formatOutput($we_doc->getElementS($regs[1]), $tag_name, $this->exportType, 2, $cdata);
								$usedFields[] = $tag_name;

								// Remove tagname from array
								if($records){
									$records = $this->remove_from_check_array($records, $tag_name);
								}
							}
						}
					} else {
						$elem = $we_doc->getElementS($k);
						if(we_base_util::isSerialized($elem) && is_array(($uns=we_unserialize($elem)))){ // is a we:link field
							$tag_name = self::correctTagname($k, 'link', $tag_counter);
							$content .= $this->formatOutput(self::formatOutput($we_doc->getFieldLink($uns), '', 'cdata'), $tag_name, $this->exportType, 2, $cdata);
							$usedFields[] = $tag_name;

							// Remove tagname from array
							if($records){
								$records = $this->remove_from_check_array($records, $tag_name);
							}
						} else { // is a normal text field
							$tag_name = self::correctTagname($k, 'text', $tag_counter);
							$content .= $this->formatOutput(we_document_base::parseInternalLinks($elem, $we_doc->ParentID), $tag_name, $this->exportType, 2, $cdata, $this->exportType === we_exim_ExIm::TYPE_XML);
							$usedFields[] = $tag_name;

							// Remove tagname from array
							if($records){
								$records = $this->remove_from_check_array($records, $tag_name);
							}
						}
					}

					break;
			}
		}

		foreach($records as $cur){
			'@phan-var string $cur';
			$content .= $this->formatOutput('', $cur, $this->exportType, 2, $cdata, $this->exportType === we_exim_ExIm::TYPE_XML);
			$usedFields[] = $cur;
		}

		$this->checkWriteFieldNames($usedFields);

		return fwrite($this->fh, $content) === strlen($content);
	}

	// FIXME: move this to some new class we_exim_contentProviderNonWE
	protected function object2nonWE(we_contents_objectFile $we_obj, bool $cdata = false): bool{
		if(!$this->fh){
			return false;
		}
		$content = '';

		$tableInfo_sorted = $we_obj->getSortedTableInfo($we_obj->TableID, true, DB_WE::$inst);

		$fields = $regs = [];
		foreach($tableInfo_sorted as $cur){
			// bugfix 8141
			if(preg_match('/(.+?)_(.*)/', $cur['name'], $regs)){
				$fields[] = ['name' => $regs[2], 'type' => $regs[1]];
			}
		}

		$usedFields = [];
		foreach($fields as $i => $field){
			switch($field['type']){
				case we_contents_object::TYPE_OBJECT:
				case we_contents_object::TYPE_IMG:
				case we_contents_object::TYPE_BINARY:
				case (in_array($this->exportType, [we_exim_ExIm::TYPE_CSV, we_exim_ExIm::TYPE_XML], true) ? 'OF' : ''):
					continue 2;
				default:
					//$realName = $field['type'] . '_' . $field['name'];
					$element = $we_obj->getElementByType($field['name'], $field['type']);

					$tag_name = self::correctTagname($field['name'], 'value', $i);
					$usedFields[] = $field['name'];
					$content .= $this->formatOutput(we_document_base::parseInternalLinks($element, 0), $tag_name, we_exim_ExIm::TYPE_XML, 2, $cdata, (!in_array($field['type'], [
							'date', 'int', 'float'], true)));
			}
		}

		$this->checkWriteFieldNames($usedFields);

		return fwrite($this->fh, $content) > 0;
	}

	abstract protected function formatOutput(string $content, string $tagname = '', string $format = we_exim_ExIm::TYPE_XML, int $tabs = 2, bool $cdata = false, bool $fix_content = false): string;

	private function getSelectedItems(string $selection, string $exportType, string $xmlTable, string $selectionType, int $docType, int $tableID, array $categorys, int $folder, array &$finalDocs, array &$finalTempl, array &$finalObjs, array &$finalClasses): void{
		$this->db = new DB_WE();
		switch($selection){
			case 'manual':
				switch($exportType){
					case we_exim_ExIm::TYPE_CSV:
					case we_exim_ExIm::TYPE_XML:
						switch(addTblPrefix($xmlTable)){
							case FILE_TABLE:
								$finalDocs = self::getIDs($finalDocs, FILE_TABLE);
								$finalObjs = [];
								break;
							default:
								$finalObjs = self::getIDs($finalObjs, OBJECT_FILES_TABLE);
								$finalDocs = [];
								break;
						}
						break;
					case we_exim_ExIm::TYPE_WE:
					default:
						$finalDocs = array_unique(self::getIDs($finalDocs, FILE_TABLE, false));
						$finalTempl = array_unique(self::getIDs($finalTempl, TEMPLATES_TABLE, false));
						$finalObjs = array_unique(self::getIDs($finalObjs, OBJECT_FILES_TABLE, false));
						$finalClasses = array_unique(self::getIDs($finalClasses, OBJECT_TABLE, false));
				}
				break;
			case 'auto':
			default:
				$doctype_where = '';

				switch($selectionType){
					case self::SELECTIONTYPE_CLASSNAME:
						if(we_modules_info::isActive(we_modules_info::OBJECT)){
							$finalObjs = $this->db->getAllq('SELECT ID FROM ' . OBJECT_FILES_TABLE . ' `of` WHERE `of`.IsFolder=0 AND `of`.TableID=' . $tableID . ' ' . ($categorys ? ' AND ' . we_category_category::getCatSQLTail([
									], 'of', true, 'Category', $categorys) : '') . we_exim_ExIm::queryForAllowed(OBJECT_FILES_TABLE), true);
						}
						break;
					case self::SELECTIONTYPE_DOCTYPE:
						$doctype_where = ' AND f.DocType="' . $docType . '"';
						$folder = 0;
					/* fall through */
					case self::SELECTIONTYPE_DOCUMENT:
					case FILE_TABLE:
					default:
						$ws_where = ($folder ? ' AND (f.Path LIKE "' . $this->db->escape(id2path($folder, FILE_TABLE)) . '/%" OR f.ID="' . $folder . '")' : '');

						$finalDocs = $this->db->getAllq('SELECT distinct ID FROM ' . FILE_TABLE . ' f WHERE 1 ' . $ws_where . $doctype_where . ($categorys ? ' AND ' . we_category_category::getCatSQLTail([
								], 'f', true, 'Category', $categorys) : ''), true);
						break;
				}
		}

		foreach($finalDocs as $v){
			$ct = fS('SELECT ContentType FROM ' . FILE_TABLE . ' WHERE ID=' . intval($v));

			if(empty($this->permittedContentTypes) || in_array($ct, $this->permittedContentTypes, true)){
				$this->RefTable->add(null, [
					'ID' => $v,
					'ContentType' => $ct,
					'level' => 0
					]
				);
			}
		}

		foreach($finalTempl as $v){
			if(empty($this->permittedContentTypes) || in_array(we_base_ContentTypes::TEMPLATE, $this->permittedContentTypes, true)){
				$this->RefTable->add(null, [
					'ID' => $v,
					'ContentType' => we_base_ContentTypes::TEMPLATE,
					'level' => 0
					]
				);
			}
		}

		if(is_array($finalObjs)){
			foreach($finalObjs as $v){
				if(empty($this->permittedContentTypes) || in_array(we_base_ContentTypes::OBJECT_FILE, $this->permittedContentTypes, true)){
					$this->RefTable->add(null, [
						'ID' => $v,
						'ContentType' => we_base_ContentTypes::OBJECT_FILE,
						'level' => 0
						]
					);
				}
			}
		}

		if(is_array($finalClasses)){
			if(empty($this->permittedContentTypes) || in_array(we_base_ContentTypes::OBJECT, $this->permittedContentTypes, true)){
				foreach($finalClasses as $v){
					$this->RefTable->add(null, [
						'ID' => $v,
						'ContentType' => we_base_ContentTypes::OBJECT,
						'level' => 0
					]);
				}
			}
		}
	}

	protected function remove_from_check_array(array $check_array, string $tagname = ''): array{
		return $check_array; // to be overridden
	}

	protected function checkWriteFieldNames(array $fields): void{
		// to be overridden
	}

	protected static function getIDs(array $selIDs, string $table, bool $with_dirs = false): array{
		$tmp = [];
		$db = new DB_WE();
		$allow = we_exim_ExIm::queryForAllowed($table);
		if($selIDs){
			$folders = $db->getAllq('SELECT ID FROM ' . $table . ' WHERE ID IN (' . implode(',', $selIDs) . ') AND IsFolder=1', true);
			'@phan-var int[] $folders';
			foreach($selIDs as $v){
				if($v){
					if(in_array($v, $folders)){
						if($with_dirs){
							$tmp[] = $v;
						}
					} else {
						$tmp[] = $v;
					}
				}
			}
			if($folders){
				we_base_util::readChilds($folders, $tmp, $table, false, $allow);
			}
		}
		if($with_dirs){
			return $tmp;
		}

		return $db->getAllq('SELECT ID FROM ' . $db->escape($table) . ' WHERE IsFolder=0 AND ID IN(' . ($tmp ? implode(',', $tmp) : '0') . ')', true);
	}

	protected static function correctTagname(string $tagname, string $alternative_name, int $alternative_number = -1): string{
		if($tagname != ''){
			// Remove spaces + special characters
			$tagname = preg_replace(['/\40+/', '/[^a-zA-Z0-9_]+/'], ["_", ''], $tagname);
		}

		// Set alternative name if no name is now present present
		return ($tagname ?:
			(($alternative_number != -1) ? $alternative_name . $alternative_number : $alternative_name) );
	}

	protected static function exportInfoMap(array $info): string{
		$out = '<' . self::INFO_TAG . '>';
		foreach($info as $inf){
			$out .= '<' . self::MAP_TAG;
			foreach($inf as $key => $value){
				$out .= ' ' . $key . '="' . $value . '"';
			}
			$out .= '></' . self::MAP_TAG . '>';
		}
		$out .= '</' . self::INFO_TAG . '>' .
			we_backup_util::backupMarker . "\n";
		return $out;
	}

	protected static function isInitialized(): bool{
		return !empty($_SESSION['weS']['ExImRefTable']);
	}

	// FIXME: make array instead of these single entries
	public function loadPreserves(): void{
		parent::loadPreserves();
		if(isset($_SESSION['weS']['ExImOptions'])){
			$this->options = $_SESSION['weS']['ExImOptions'];
			$this->exportType = $_SESSION['weS']['ExImExportType'];
			$this->exportProperties = $_SESSION['weS']['ExImProperties'];
		}
	}

	public function savePreserves(): void{
		parent::savePreserves();
		$_SESSION['weS']['ExImOptions'] = $this->options;
		$_SESSION['weS']['ExImExportType'] = $this->exportType;
		$_SESSION['weS']['ExImProperties'] = $this->exportProperties;
	}

	public static function unsetPreserves(): void{
		parent::unsetPreserves();
		if(isset($_SESSION['weS']['ExImOptions'])){
			unset($_SESSION['weS']['ExImOptions'],
				 $_SESSION['weS']['ExImExportType'],
				 $_SESSION['weS']['ExImProperties']);
		}
	}

}
