<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: e818c44d0fb7 $
 * $Author: Lukas Imhof $
 * $Date: 2019-10-21 17:28 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Filename:    we_html_select.inc.php
 *
 * Function:    Utility class that implements operations on selects
 *
 * Description: Provides functions for creating html tags
 */
class we_html_select extends we_html_baseCollection{

	/**
	 * Constructor
	 *
	 * @param		$attribs								array			(optional)
	 *
	 */
	public function __construct(array $attribs = []){
		parent::__construct('select', true, $attribs);
	}

	/**
	 * The function returns number of options
	 *
	 * @return		int
	 */
	function getOptionNum(): int{
		return (count($this->childs) - 1);
	}

	/**
	 * The function returns select options as array of we_html_baseElement
	 *
	 * @return		array of we_html_baseElement
	 */
	public function getOptionsArray(): array{
		return $this->childs;
	}

	/**
	 * The function add new option to a select box
	 *
	 * @param		$value									string
	 * @param		$text									string
	 *
	 * @return		we_html_select
	 */
	public function addOption(string $value, string $text, array $attribs = []): self{
		$attribs['value'] = $value;
		$this->childs[] = new we_html_baseElement('option', true, $attribs, $text);
		return $this;
	}

	/**
	 * The function adds one or more options to a select box
	 * @param array<string|int,string|int> $entries
	 * @return		we_html_select
	 */
	public function addOptions(array $entries = []): self{
		foreach($entries as $value => $text){
			$this->childs[] = new we_html_baseElement('option', true, ['value' => (string) $value], (string) $text);
		}
		return $this;
	}

	/**
	 * The function deletes option with given option value
	 *
	 * @param		$value										string
	 * @return		we_html_select
	 */
	public function delOption(string $value): self{
		$cid = -1;
		foreach($this->childs as $k => $v){
			if($v->attribs['value'] === $value){
				$cid = (int) $k;
				break;
			}
		}
		if($cid !== -1){
			$this->delChild($cid);
		}
		return $this;
	}

	/**
	 * The function deletes all options from select
	 *
	 * @return		we_html_select
	 */
	public function delAllOptions(): self{
		$this->childs = [];

		return $this;
	}

	/**
	 * The function inserts option on specified place in a select box
	 * Parameter $optid defines option's place in select's child array
	 * If $over is true then an option which is placed on the specified position will be overwritten
	 *
	 * @param		$optid									int
	 * @param		$value									string
	 * @param		$text									string
	 * @param		$over									string			(optional)
	 * @return		we_html_select
	 */
	public function insertOption(int $optid, string $value, string $text, bool $over = false): self{
		$new_opt = new we_html_baseElement('option', true, ['value' => $value], $text);

		if($over){
			$this->childs[$optid] = $new_opt;
		} else {
			if($optid == 0){
				$optid = -1;
			}
			if(count($this->childs) >= $optid + 1){
				$array_pre = array_slice($this->childs, 0, ($optid + 1));
				$array_pre[] = $new_opt;
				$array_post = array_slice($this->childs, ($optid + 1));
				$this->childs = array_merge($array_pre, $array_post);
			} else {
				$this->childs[] = $new_opt;
			}
		}

		return $this;
	}

	/**
	 * The function sets option attributes and content. The option is identified by optid.
	 *
	 * @param		int $optid
	 * @param		array<string,string|string[]> $attribs
	 * @return		we_html_select
	 */
	public function setOption(int $optid, array $attribs = [], string $content = ''): self{
		$opt = & $this->getChild($optid);
		$opt->setAttributes($attribs);
		if($content !== ''){
			$this->content = $content;
		}

		return $this;
	}

	/**
	 * The function selects option that is identified by the value.
	 *
	 * @param		$value									string
	 * @return		we_html_select
	 */
	public function selectOption($value): self{
		if(!isset($this->attribs['multiple'])){
			$this->unselectAllOptions();
		}
		foreach($this->childs as $k => $v){
			//fix #7912
			$equal = (is_bool($v->attribs['value']) || is_bool($value)) ? $v->attribs['value'] == $value :
				(string) $v->attribs['value'] == (string) $value;

			if($equal){
				$this->setOption((int) $k, ['selected' => 'selected']);
				break;
			}
		}

		return $this;
	}

	/**
	 * The function unsets all selected options
	 *
	 * @return		we_html_select
	 */
	public function unselectAllOptions(): self{
		foreach($this->childs as $v){
			unset($v->attribs['selected']);
		}

		return $this;
	}

	/**
	 * The function sets option identified by optid with given value and text
	 * @param		$optid									int
	 * @param		$value									string
	 * @param		$text									string
	 *
	 * @return		we_html_select
	 */
	function setOptionVT(int $optid, string $value, string $text): self{
		$opt = & $this->getChild($optid);
		$opt->setAttribute('value', $value);
		$this->content = $text;

		return $this;
	}

	/**
	 * The function adds a new option group to the select box
	 *
	 * @param  $attribs        array
	 * @return  we_html_select
	 */
	public function addOptionGroup(array $attribs = []): self{
		$this->childs[] = new we_html_baseCollection('optgroup', true, $attribs);

		return $this;
	}

	/**
	 * The function returns a new option. This function is static.
	 *
	 * @param  $value         string
	 * @param  $text         string
	 * @return  we_html_baseElement
	 */
	public static function getNewOption(string $value, string $text): we_html_baseElement{
		return new we_html_baseElement('option', true, ['value' => $value], $text);
	}

	/**
	 * The function returns a new option group. This function is static.
	 *
	 * @param  $attribs        array
	 *
	 * @return  we_html_baseCollection
	 */
	public static function getNewOptionGroup(array $attribs = []): we_html_baseCollection{
		return new we_html_baseCollection('optgroup', true, $attribs);
	}

}
